import mimetypes
import mmap
import os
import platform

from contextlib import contextmanager
from pathlib import Path

from .conf import settings


class SourceFile:

    module_mime_types = [
        "application/javascript"
    ]
    module_detect_bytes = [
        b'export ',
        b'import ',
    ]
    rollup_hints = [
        b'"use rollup"',
        b"'use rollup'",
    ]

    def __init__(self, source_path):
        if isinstance(source_path, Path):
            self.source_path = source_path
            self.source_path_str = str(source_path)
        elif isinstance(source_path, str):
            self.source_path = Path(source_path)
            self.source_path_str = source_path
        else:
            raise ValueError('Invalid source_path value, should be Path or str')
        self.stat = self.source_path.stat()
        content_type, self.encoding = mimetypes.guess_type(self.source_path_str)
        self.content_type = content_type or 'application/octet-stream'

    def is_rollup_module(self, mfile):
        if self.content_type in self.module_mime_types:
            return any([mfile.find(b) != -1 for b in self.module_detect_bytes])
        else:
            return False

    def has_to_import(self):
        if settings.DENO_ENABLE and self.stat.st_size > 0:
            with self.get_mmap() as mfile:
                return self.is_rollup_module(mfile)
        else:
            return False

    @contextmanager
    def get_mmap(self):
        try:
            fd = None
            mfile = None
            fd = os.open(f"{self.source_path_str}", os.O_RDONLY)
            if platform.system() == 'Windows':
                mfile = mmap.mmap(fd, 0, access=mmap.ACCESS_READ)
            else:
                mfile = mmap.mmap(fd, 0, prot=mmap.PROT_READ)
            yield mfile
        finally:
            if mfile is not None:
                mfile.close()
            if fd is not None:
                os.close(fd)

    def should_rollup(self):
        if settings.DENO_ENABLE and self.stat.st_size > 0:
            with self.get_mmap() as mfile:
                if self.is_rollup_module(mfile):
                    for hint in self.rollup_hints:
                        if mfile[:len(hint)] == hint:
                            return True
                if settings.DENO_ROLLUP_MATCH_PATH:
                    should_rollup = any([
                        self.source_path_str.endswith(Path(entry_point).__str__())
                        for entry_point in settings.DENO_ROLLUP_ENTRY_POINTS
                    ])
                    return should_rollup
        return False



# From https://github.com/googlearchive/appengine-try-python-flask/blob/master/lib/werkzeug/contrib/iterio.py

def implements_iterator(cls):
    cls.next = cls.__next__
    del cls.__next__
    return cls


def _mixed_join(iterable, sentinel):
    """concatenate any string type in an intelligent way."""
    iterator = iter(iterable)
    first_item = next(iterator, sentinel)
    if isinstance(first_item, bytes):
        return first_item + b''.join(iterator)
    return first_item + u''.join(iterator)


def _newline(reference_string):
    if isinstance(reference_string, bytes):
        return b'\n'
    return u'\n'


@implements_iterator
class IterIO(object):
    """Instances of this object implement an interface compatible with the
    standard Python :class:`file` object.  Streams are either read-only or
    write-only depending on how the object is created.
    If the first argument is an iterable a file like object is returned that
    returns the contents of the iterable.  In case the iterable is empty
    read operations will return the sentinel value.
    If the first argument is a callable then the stream object will be
    created and passed to that function.  The caller itself however will
    not receive a stream but an iterable.  The function will be be executed
    step by step as something iterates over the returned iterable.  Each
    call to :meth:`flush` will create an item for the iterable.  If
    :meth:`flush` is called without any writes in-between the sentinel
    value will be yielded.
    Note for Python 3: due to the incompatible interface of bytes and
    streams you should set the sentinel value explicitly to an empty
    bytestring (``b''``) if you are expecting to deal with bytes as
    otherwise the end of the stream is marked with the wrong sentinel
    value.
    .. versionadded:: 0.9
       `sentinel` parameter was added.
    """

    def __new__(cls, obj, sentinel=''):
        try:
            iterator = iter(obj)
        except TypeError:
            raise NotImplementedError("IterI is not implemented")
        return IterO(iterator, sentinel)

    def __iter__(self):
        return self

    def tell(self):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        return self.pos

    def isatty(self):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        return False

    def seek(self, pos, mode=0):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        raise IOError(9, 'Bad file descriptor')

    def truncate(self, size=None):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        raise IOError(9, 'Bad file descriptor')

    def write(self, s):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        raise IOError(9, 'Bad file descriptor')

    def writelines(self, list):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        raise IOError(9, 'Bad file descriptor')

    def read(self, n=-1):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        raise IOError(9, 'Bad file descriptor')

    def readlines(self, sizehint=0):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        raise IOError(9, 'Bad file descriptor')

    def readline(self, length=None):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        raise IOError(9, 'Bad file descriptor')

    def flush(self):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        raise IOError(9, 'Bad file descriptor')

    def __next__(self):
        if self.closed:
            raise StopIteration()
        line = self.readline()
        if not line:
            raise StopIteration()
        return line


class IterO(IterIO):
    """Iter output.  Wrap an iterator and give it a stream like interface."""

    def __new__(cls, gen, sentinel=''):
        self = object.__new__(cls)
        self._gen = gen
        self._buf = None
        self.sentinel = sentinel
        self.closed = False
        self.pos = 0
        return self

    def __iter__(self):
        return self

    def _buf_append(self, string):
        '''Replace string directly without appending to an empty string,
        avoiding type issues.'''
        if not self._buf:
            self._buf = string
        else:
            self._buf += string

    def close(self):
        if not self.closed:
            self.closed = True
            if hasattr(self._gen, 'close'):
                self._gen.close()

    def seek(self, pos, mode=0):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        if mode == 1:
            pos += self.pos
        elif mode == 2:
            self.read()
            self.pos = min(self.pos, self.pos + pos)
            return
        elif mode != 0:
            raise IOError('Invalid argument')
        buf = []
        try:
            tmp_end_pos = len(self._buf)
            while pos > tmp_end_pos:
                item = next(self._gen)
                tmp_end_pos += len(item)
                buf.append(item)
        except StopIteration:
            pass
        if buf:
            self._buf_append(_mixed_join(buf, self.sentinel))
        self.pos = max(0, pos)

    def read(self, n=-1):
        if self.closed:
            raise ValueError('I/O operation on closed file')
        if n < 0:
            self._buf_append(_mixed_join(self._gen, self.sentinel))
            result = self._buf[self.pos:]
            self.pos += len(result)
            return result
        new_pos = self.pos + n
        buf = []
        try:
            tmp_end_pos = 0 if self._buf is None else len(self._buf)
            while new_pos > tmp_end_pos or (self._buf is None and not buf):
                item = next(self._gen)
                tmp_end_pos += len(item)
                buf.append(item)
        except StopIteration:
            pass
        if buf:
            self._buf_append(_mixed_join(buf, self.sentinel))

        if self._buf is None:
            return self.sentinel

        new_pos = max(0, new_pos)
        try:
            return self._buf[self.pos:new_pos]
        finally:
            self.pos = min(new_pos, len(self._buf))

    def readline(self, length=None):
        if self.closed:
            raise ValueError('I/O operation on closed file')

        nl_pos = -1
        if self._buf:
            nl_pos = self._buf.find(_newline(self._buf), self.pos)
        buf = []
        try:
            pos = self.pos
            while nl_pos < 0:
                item = next(self._gen)
                local_pos = item.find(_newline(item))
                buf.append(item)
                if local_pos >= 0:
                    nl_pos = pos + local_pos
                    break
                pos += len(item)
        except StopIteration:
            pass
        if buf:
            self._buf_append(_mixed_join(buf, self.sentinel))

        if self._buf is None:
            return self.sentinel

        if nl_pos < 0:
            new_pos = len(self._buf)
        else:
            new_pos = nl_pos + 1
        if length is not None and self.pos + length < new_pos:
            new_pos = self.pos + length
        try:
            return self._buf[self.pos:new_pos]
        finally:
            self.pos = min(new_pos, len(self._buf))

    def readlines(self, sizehint=0):
        total = 0
        lines = []
        line = self.readline()
        while line:
            lines.append(line)
            total += len(line)
            if 0 < sizehint <= total:
                break
            line = self.readline()
        return lines



from .deno import DenoApi


class DenoMaps(DenoApi):

    location = '/maps/'
    response_post_schema = {
        'type': 'object',
        'properties': {
            'server': {'type': 'string'},
            'pid': {'type': 'integer'},
            'version': {'type': 'string'},
        }
    }



from .run import DenoRun

from ..conf.settings import DENO_SERVER


class DenoServer(DenoRun):

    script_name = "server.ts"
    script_args = [
        f"--host={DENO_SERVER['hostname']}",
        f"--port={DENO_SERVER['port']}"
    ]



from collections import MutableMapping
import os

from pathlib import Path

from django.apps import apps
from django.conf import settings
from django.contrib.staticfiles.finders import get_finders
from django.contrib.staticfiles.storage import staticfiles_storage
from django.utils.functional import cached_property

from .sourcefile import SourceFile


# Todo check with Django 2.2 / Django 3.2.
"""
import os
from django.conf import settings
from django_deno.importmap import ImportMapGenerator
import_map_generator = ImportMapGenerator()
import_map = import_map_generator.get_import_map(os.path.join(settings.BASE_DIR, 'drf_gallery', 'static', 'components', 'main.js'))
serialized_map_generator = import_map_generator.serialize()
import_map_generator = ImportMapGenerator(cache_entry=serialized_map_generator)
serialized_map_generator == import_map_generator.serialize()
import_map == import_map_generator.get_import_map(os.path.join(settings.BASE_DIR, 'drf_gallery', 'static', 'components', 'main.js'))
"""


class CommonBasePath:

    def __init__(self, cache_entry=None):
        if cache_entry is None:
            self.create()
        else:
            self.deserialize(cache_entry)

    def __eq__(self, other):
        return self._common_base == other._common_base and self.common_base_str == other.common_base_str

    def __str__(self):
        return os.path.join(*(self.common_base + ['']))

    @property
    def common_base(self):
        return self._common_base

    @common_base.setter
    def common_base(self, path_parts):
        self._common_base = path_parts
        self.common_base_str = str(self)

    def create(self):
        # dict with path parts
        self.common_base = []
        # path str
        self.common_base_str = ''

    def serialize(self):
        return str(self)

    def deserialize(self, cache_entry: str):
        path_obj = Path(cache_entry)
        self.common_base = list(path_obj.parts)

    def __repr__(self):
        return f"{type(self).__name__}({self.common_base})"

    def split_parents(self, path_obj):
        return list(path_obj.parts)[:-1]

    def apply_path(self, path):
        path_obj = Path(path)
        if len(self.common_base) == 0:
            self.common_base = self.split_parents(path_obj)
        else:
            path_parents = self.split_parents(path_obj)
            for idx, common_parent in enumerate(self.common_base):
                if path_parents[idx] != common_parent:
                    self.common_base = self.common_base[:idx]
                    break

    def pack(self, path_str):
        if path_str.startswith(self.common_base_str):
            return path_str[len(self.common_base_str):]
        else:
            return path_str

    def unpack(self, path_str):
        if path_str.startswith(self.common_base_str):
            return path_str
        else:
            return f"{self.common_base_str}{path_str}"


class PathMap(MutableMapping):

    def __init__(self, cache_entry=None):
        if cache_entry is None:
            self.create()
        else:
            self.deserialize(cache_entry)

    def __eq__(self, other):
        return self.map == other.map and self.base_key == other.base_key and self.base_val == other.base_val

    def create(self):
        self.map = {}
        self.base_key = CommonBasePath()
        self.base_val = CommonBasePath()

    def __getitem__(self, k):
        return self.map[k]

    # Warning: Assumes k:v are unpacked. Automatically begins to unpack k:v.
    # Do not call after .pack(), otherwise one have to perform .unpack() / .pack() later.
    def __setitem__(self, k, v):
        self.map[k] = v
        self.base_key.apply_path(k)
        self.base_val.apply_path(v)

    def __delitem__(self, k):
        raise NotImplementedError('Operation is not supported')

    def __iter__(self):
        return iter(self.map)

    def __len__(self):
        return len(self.map)

    def __repr__(self):
        return f"{type(self).__name__}({self.map})"

    def unpack_relation(self, rel_k, rel_v):
        if rel_v == '':
            packed_v = rel_k
        elif rel_v.endswith('.'):
            packed_k_no_rel = rel_k.lstrip('.' + os.sep)
            packed_v = f"{rel_v[:-1]}{packed_k_no_rel}"
        else:
            packed_v = rel_v
        return rel_k, packed_v

    def pack_relation(self, packed_k, packed_v):
        packed_k_no_rel = packed_k.lstrip('.' + os.sep)
        if packed_k == packed_v:
            rel_v = ''
        elif packed_v.endswith(packed_k_no_rel):
            rel_v = packed_v[:-len(packed_k_no_rel)] + '.'
        else:
            rel_v = packed_v
        return packed_k, rel_v

    def get_unpacked(self, k):
        # Assuming that map is currently packed
        rel_k = self.base_key.pack(k)
        packed_k, packed_v = self.unpack_relation(rel_k, self.map[rel_k])
        return self.base_val.unpack(packed_v)

    def unpack_items(self):
        for rel_k, rel_v in self.map.items():
            packed_k, packed_v = self.unpack_relation(rel_k, rel_v)
            yield self.base_key.unpack(packed_k), self.base_val.unpack(packed_v)

    def pack(self):
        map = {}
        for k, v in self.map.items():
            packed_k = self.base_key.pack(k)
            packed_v = self.base_val.pack(v)
            rel_k, rel_v = self.pack_relation(packed_k, packed_v)
            map[rel_k] = rel_v
        self.map = map

    def serialize(self):
        return {
            'map': self.map,
            'base_key': self.base_key.serialize(),
            'base_val': self.base_val.serialize(),
        }

    def deserialize(self, cache_entry):
        self.map = cache_entry['map']
        self.base_key = CommonBasePath(cache_entry=cache_entry['base_key'])
        self.base_val = CommonBasePath(cache_entry=cache_entry['base_val'])

    def __str__(self):
        return str(self.serialize())


class ImportMapGenerator:

    def __init__(self, cache_entry=None, logger=None):
        self.logger = logger
        if cache_entry is None:
            self.create()
        else:
            self.deserialize(cache_entry)

    def log(self, msg, **kwargs):
        if self.logger is not None:
            self.logger.write(msg)
        else:
            raise ValueError(msg)

    def create(self):
        self.module_basedir = None
        self.storage = staticfiles_storage
        if not self.local:
            raise ValueError('Only local storage is supported')
        self.ignore_patterns = apps.get_app_config('staticfiles').ignore_patterns
        self.mapped_files = set()
        self.base_map = PathMap()
        self.import_map = PathMap()
        self.collect_import_map()
        self.base_map.pack()
        self.import_map.pack()

    # django.contrib.staticfiles.management.commands.collectstatic.Command.local
    @cached_property
    def local(self):
        try:
            self.storage.path('')
        except NotImplementedError:
            return False
        return True

    def collect_import_map(self):
        found_files = {}
        for finder in get_finders():
            for path, storage in finder.list(self.ignore_patterns):
                # Prefix the relative path if the source storage contains it
                if getattr(storage, 'prefix', None):
                    prefixed_path = os.path.join(storage.prefix, path)
                else:
                    prefixed_path = path

                if prefixed_path not in found_files:
                    found_files[prefixed_path] = (storage, path)
                    self.add_to_import_map(path, prefixed_path, storage)
                else:
                    self.log(
                        "Found another file with the destination path '%s'. It "
                        "will be ignored since only the first encountered file "
                        "is collected. If this is not what you want, make sure "
                        "every static file has a unique path." % prefixed_path,
                        level=1,
                    )

    def add_to_import_map(self, path, prefixed_path, source_storage):
        if prefixed_path not in self.mapped_files:
            self.mapped_files.add(prefixed_path)
            # self.storage.exists(prefixed_path)
            # When was the target file modified last time?
            # target_last_modified = self.storage.get_modified_time(prefixed_path)
            # When was the source file modified last time?
            # source_last_modified = source_storage.get_modified_time(path)
            # Avoid sub-second precision (see #14665, #19540)
            # file_is_unmodified = (
            #         target_last_modified.replace(microsecond=0) >=
            #         source_last_modified.replace(microsecond=0)
            # )
            # Then delete the existing file if really needed
            # self.storage.delete(prefixed_path)
            # The full path of the source file
            source_path = source_storage.path(path)
            # The full path of the target file
            target_path = self.storage.path(prefixed_path)
            source_file = SourceFile(source_path)
            if source_file.has_to_import():
                if source_path.startswith(settings.BASE_DIR):
                    self.base_map[source_path] = target_path
                else:
                    self.import_map[target_path] = source_path

    def has_common_path(self, path):
        try:
            common_path = os.path.commonpath([self.module_basedir, path])
        except ValueError:
            return False
        return True

    def to_relative_path(self, path):
        relative_path = os.path.relpath(path, self.module_basedir)
        return relative_path

    # es_module_path - a full path to valid existing es module
    # warning: called after .deserialize(), thus only .base_map / .import_map instance attributes are
    # assumed to be properly initialized.
    def get_import_map(self, es_module_path):
        self.module_basedir = os.path.dirname(
            self.base_map.get_unpacked(es_module_path)
        )
        relative_import_map = PathMap()
        for target_path, source_path in self.import_map.unpack_items():
            if self.has_common_path(source_path):
                relative_target_path = self.to_relative_path(target_path)
                relative_import_map[relative_target_path] = source_path
        relative_import_map.pack()
        return relative_import_map

    def serialize(self):
        return {
            'base_map': self.base_map.serialize(),
            'import_map': self.import_map.serialize(),
        }

    def deserialize(self, cache_entry):
        self.base_map = PathMap(cache_entry=cache_entry['base_map'])
        self.import_map = PathMap(cache_entry=cache_entry['import_map'])



import os

from django.core.exceptions import ImproperlyConfigured
from django.conf import settings

DENO_ENABLE = getattr(settings, 'DENO_ENABLE', True)
DENO_RELOAD = getattr(settings, 'DENO_RELOAD', False)
DENO_CHECK_LOCK_FILE = getattr(settings, 'DENO_CHECK_LOCK_FILE', False)
DENO_USE_VENDOR = getattr(settings, 'DENO_USE_VENDOR', True)
DENO_ROLLUP_MATCH_PATH = getattr(settings, 'DENO_ROLLUP_MATCH_PATH', True)

DENO_INSTALL = getattr(settings, 'DENO_INSTALL', os.getenv('DENO_INSTALL'))
if DENO_INSTALL is None:
    raise ImproperlyConfigured('Please set DENO_INSTALL environment variable or project settings.DENO_INSTALL path')

DENO_PATH = os.path.join(DENO_INSTALL, 'bin', 'deno')

if getattr(settings, 'DENO_DEBUG', False):
    DENO_TIMEOUT = 60
else:
    DENO_TIMEOUT = 20

DENO_SERVER = {
    'scheme': 'http',
    'hostname': '127.0.0.1',
    'port': 8099,
}
DENO_SERVER.update(getattr(settings, 'DENO_SERVER', {}))

DENO_URL = f'{DENO_SERVER["scheme"]}://{DENO_SERVER["hostname"]}:{DENO_SERVER["port"]}'

DENO_PROXY_CHUNK_SIZE = getattr(settings, 'DENO_PROXY_CHUNK_SIZE', 256 * 1024)

# Map of rollup.js output module type to html script tag module type:
DENO_OUTPUT_MODULE_FORMATS = {
    'module': 'es',
    'systemjs-module': 'system',
}
DENO_OUTPUT_MODULE_FORMATS.update(getattr(settings, 'DENO_OUTPUT_MODULE_FORMATS', {}))

DENO_ROLLUP_SERVE_OPTIONS = {
    'inlineFileMap': True,
    'relativePaths': True,
    'preserveEntrySignatures': False,
    'staticFilesResolver': True,
    'withCache': True,
}
DENO_ROLLUP_SERVE_OPTIONS.update(getattr(settings, 'DENO_ROLLUP_SERVE_OPTIONS', {}))

DENO_OUTPUT_MODULE_TYPE = getattr(settings, 'DENO_OUTPUT_MODULE_TYPE', 'module')

DENO_ROLLUP_COLLECT_OPTIONS = {
    # 'relativePaths': True,
    # 'staticFilesResolver': True,
    'terser': True,
    'bundles': getattr(settings, 'DENO_ROLLUP_BUNDLES', {}),
    'moduleFormat': DENO_OUTPUT_MODULE_FORMATS[DENO_OUTPUT_MODULE_TYPE],
}
DENO_ROLLUP_COLLECT_OPTIONS.update(getattr(settings, 'DENO_ROLLUP_COLLECT_OPTIONS', {}))

DENO_ROLLUP_ENTRY_POINTS = getattr(settings, 'DENO_ROLLUP_ENTRY_POINTS', [])



import json
import os
import re
from urllib.parse import urlsplit, urlunsplit

from .base import ExecDeno, DENO_SCRIPT_PATH

from ..regex import finditer_with_separators, MatchGroup

from ..conf.settings import DENO_RELOAD, DENO_CHECK_LOCK_FILE, DENO_USE_VENDOR


class DenoRun(ExecDeno):

    deno_command = 'run'
    deno_flags = ["-A", "--unstable", "--allow-net"]

    # https://regexr.com
    module_version_split = re.compile(r'\/[^\/]+@.+?\/')
    module_version_replace = re.compile(r'@.+?\/')

    def generate_importmap(self):
        deno_import_map = {}
        with open(self.deno_lock_path, "r") as deno_lock_file:
            deno_lock = json.load(deno_lock_file)
            for specific_url in deno_lock.keys():
                url_parts = urlsplit(specific_url)._asdict()
                version_parts = finditer_with_separators(self.module_version_split, url_parts['path'])
                for i, version_part in enumerate(version_parts):
                    if isinstance(version_part, MatchGroup):
                        version_parts[i] = self.module_version_replace.sub('/', version_part)
                        url_parts['path'] = ''.join(version_parts)
                        unspecific_url = urlunsplit(url_parts.values())
                        deno_import_map[unspecific_url] = specific_url
                        break
            if len(deno_import_map) > 0:
                with open(self.run_importmap_path, "w+") as deno_import_map_file:
                    json.dump({'imports': deno_import_map}, deno_import_map_file, indent=2, ensure_ascii=False)

    # return true if importmap exists, false otherwise
    def cache_importmap(self):
        try:
            deno_importmap_mtime = os.path.getmtime(self.run_importmap_path)
        except OSError:
            deno_importmap_mtime = None
        try:
            if deno_importmap_mtime is None or os.path.getmtime(self.deno_lock_path) > deno_importmap_mtime:
                self.generate_importmap()
                return True
            else:
                return deno_importmap_mtime is not None
        except OSError:
            return False

    def get_deno_flags(self):
        deno_flags = super().get_deno_flags()
        if DENO_USE_VENDOR:
            deno_flags.extend([
                "--no-remote",
                f"--import-map={self.vendor_importmap_path}"
            ])
        else:
            if DENO_RELOAD and DENO_CHECK_LOCK_FILE:
                raise ValueError("DENO_RELOAD / DENO_CHECK_LOCK_FILE are mutually exclusive options")
            if DENO_RELOAD:
                deno_flags.extend(["--reload", "--lock-write", f"--lock={self.deno_lock_path}"])
            if DENO_CHECK_LOCK_FILE:
                if not os.path.isfile(self.deno_lock_path):
                    raise ValueError(
                        f"Lock file is not found: {self.deno_lock_path}\n"
                        f"Please use DENO_RELOAD=True option to create the lock file first."
                    )
                deno_flags.append(f"--lock={self.deno_lock_path}")
                if self.cache_importmap():
                    deno_flags.append(f"--import-map={self.run_importmap_path}")
        return deno_flags



import signal
import threading

from django.conf import settings
from django.contrib.staticfiles.management.commands import runserver

from ...commands import DenoProcess
from ...handlers import RollupFilesHandler

lock = threading.Lock()

global deno_process
deno_process = None


class Command(runserver.Command, DenoProcess):

    help = f"{runserver.Command.help} Starts Deno Web server to serve rollup bundles from static files."

    def get_handler(self, *args, **options):
        global deno_process
        self.orig_sigint = None
        deno_process = self.run_deno_process()
        """
        Return the static files serving handler wrapping the default handler,
        if static files should be served. Otherwise return the default handler.
        """
        handler = super().get_handler(*args, **options)
        use_static_handler = options['use_static_handler']
        insecure_serving = options['insecure_serving']
        if use_static_handler and (settings.DEBUG or insecure_serving):
            return RollupFilesHandler(handler)
        return handler

    def terminate_deno(self):
        global deno_process
        with lock:
            if deno_process is not None:
                self.stdout.write(f"Terminating deno server pid={deno_process.pid}")
                deno_process.terminate()
                deno_process = None

    def sigint_handler(self, signum, frame):
        self.terminate_deno()
        if callable(self.orig_sigint):
            self.orig_sigint(signum, frame)

    def run(self, **options):
        # Run before autoreload.run_with_reloader() spawns another thread:
        if threading.main_thread() == threading.current_thread():
            self.orig_sigint = signal.getsignal(signal.SIGINT)
            signal.signal(signal.SIGINT, self.sigint_handler)
        super().run(**options)



import os
import posixpath

from pathlib import Path

from django.utils._os import safe_join
from django.utils.http import http_date
from django.utils.translation import gettext as _
from django.contrib.staticfiles import finders
from django.conf import settings
from django.views.static import directory_index, was_modified_since
from django.http import (
    FileResponse, Http404, HttpResponseNotModified, StreamingHttpResponse
)

from .sourcefile import SourceFile
from .conf import settings as deno_settings
from .api.rollup import DenoRollup


# from django.views.static import serve
def serve_rollup(request, path, document_root=None, show_indexes=False):
    """
    Serve static files below a given point in the directory structure.

    To use, put a URL pattern such as::

        from django.views.static import serve

        path('<path:path>', serve, {'document_root': '/path/to/my/files/'})

    in your URLconf. You must provide the ``document_root`` param. You may
    also set ``show_indexes`` to ``True`` if you'd like to serve a basic index
    of the directory.  This index view will use the template hardcoded below,
    but if you'd like to override it, you can create a template called
    ``static/directory_index.html``.
    """
    path = posixpath.normpath(path).lstrip('/')
    fullpath = Path(safe_join(document_root, path))
    if fullpath.is_dir():
        if show_indexes:
            return directory_index(path, fullpath)
        raise Http404(_("Directory indexes are not allowed here."))
    if not fullpath.exists():
        raise Http404(_('“%(path)s” does not exist') % {'path': fullpath})
    statobj = fullpath.stat()

    source_file = SourceFile(fullpath)
    if source_file.should_rollup():
        response = DenoRollup(content_type=source_file.content_type).post({
            'filename': str(fullpath.name),
            'basedir': str(fullpath.parent),
            'options': deno_settings.DENO_ROLLUP_SERVE_OPTIONS,
        })
        if not isinstance(response, StreamingHttpResponse):
            # report error
            return response
    else:
        # Respect the If-Modified-Since header.
        if not was_modified_since(request.META.get('HTTP_IF_MODIFIED_SINCE'),
                                  statobj.st_mtime, statobj.st_size):
            return HttpResponseNotModified()
        response = FileResponse(fullpath.open('rb'), content_type=source_file.content_type)
    response["Last-Modified"] = http_date(statobj.st_mtime)
    if source_file.encoding:
        response["Content-Encoding"] = source_file.encoding
    return response


# from django.contrib.staticfiles.views import serve
def serve(request, path, insecure=False, **kwargs):
    """
    Serve static files below a given point in the directory structure or
    from locations inferred from the staticfiles finders.

    To use, put a URL pattern such as::

        from django.contrib.staticfiles import views

        path('<path:path>', views.serve)

    in your URLconf.

    It uses the django.views.static.serve() view to serve the found files.
    """
    if not settings.DEBUG and not insecure:
        raise Http404
    normalized_path = posixpath.normpath(path).lstrip('/')
    absolute_path = finders.find(normalized_path)
    if not absolute_path:
        if path.endswith('/') or path == '':
            raise Http404("Directory indexes are not allowed here.")
        raise Http404("'%s' could not be found" % path)
    document_root, path = os.path.split(absolute_path)
    return serve_rollup(request, path, document_root=document_root, **kwargs)



import json

from django.http import (
    HttpResponse, StreamingHttpResponse
)

from .deno import DenoApi

from ..conf import settings
from ..utils import ex_to_str


class DenoRollup(DenoApi):

    location = '/rollup/'
    extra_post_kwargs = {'stream': True}

    def __init__(self, content_type, **kwargs):
        self.content_type = content_type
        super().__init__(**kwargs)

    def parse_post_response(self, response):
        if response.status_code == 200:
            response = StreamingHttpResponse(
                response.iter_content(chunk_size=settings.DENO_PROXY_CHUNK_SIZE), content_type=self.content_type
            )
        else:
            response = HttpResponse(
                'throw Error({});'.format(json.dumps(response.text)),
                content_type=self.content_type
            )
            response.status_code = 200
        return response

    def parse_not_responding_error(self, ex):
        ex_string = json.dumps(ex_to_str(ex))
        response = HttpResponse(
            f'throw Error({ex_string});',
            content_type=self.content_type
        )
        response.status_code = 200
        return response

    def parse_http_error(self, ex):
        return self.parse_not_responding_error()



from django.contrib.staticfiles.handlers import StaticFilesHandler

from .views import serve


class RollupFilesHandler(StaticFilesHandler):

    def serve(self, request):
        """Serve the request path."""
        return serve(request, self.file_path(request.path), insecure=True)



import inspect
import os
import subprocess
from copy import copy

from django.conf import settings

from ..conf.settings import DENO_PATH


class ExecDeno:

    deno_command = None
    deno_flags = []
    script_name = None
    script_args = []
    # https://deno.land/manual/linking_to_external_code/integrity_checking
    # https://deno.land/manual/linking_to_external_code
    # https://deno.land/manual/tools/vendor
    deno_lock_filename = 'lock.json'
    run_importmap_filename = 'import_map.json'

    def get_deno_flags(self):
        deno_flags = copy(self.deno_flags)
        if getattr(settings, 'DENO_DEBUG', False):
            # chrome://inspect
            deno_flags.append("--inspect-brk")
        return deno_flags

    def get_script_name(self):
        return self.script_name

    def get_deno_command(self):
        return self.deno_command

    def get_script_args(self):
        return self.script_args

    def get_shell_args(self):
        return [
            DENO_PATH, self.get_deno_command()
        ] + self.get_deno_flags() + [
            os.path.join(DENO_SCRIPT_PATH, self.get_script_name()),
        ] + self.get_script_args()

    def get_popen_kwargs(self):
        return {
            'shell': False,
        }

    def __init__(self, deno_lock_filename=None, run_importmap_filename=None, **kwargs):
        if deno_lock_filename is None:
            deno_lock_filename = self.deno_lock_filename
        self.deno_lock_path = os.path.join(DENO_SCRIPT_PATH, deno_lock_filename)

        if run_importmap_filename is None:
            run_importmap_filename = self.run_importmap_filename
        self.run_importmap_path = os.path.join(DENO_SCRIPT_PATH, run_importmap_filename)

        self.vendor_importmap_path = os.path.join(DENO_SCRIPT_PATH, 'vendor', 'import_map.json')

        self.shell_args = self.get_shell_args()

    def __str__(self):
        return ' '.join(self.shell_args)

    def __call__(self, *args, **kwargs):
        popen_kwargs = self.get_popen_kwargs()
        deno_process = subprocess.Popen(self.shell_args, **popen_kwargs)
        return deno_process


DENO_SCRIPT_PATH = os.path.join(
    os.path.dirname(os.path.dirname(inspect.getfile(ExecDeno))),
    'deno'
)



import os

from django.core.management.base import BaseCommand

from ...process.base import DENO_SCRIPT_PATH
from ...process.vendor import DenoVendor


class Command(BaseCommand):
    help = 'Generate deno vendor bundle'

    def handle(self, *args, **options):
        deno_vendor = DenoVendor()
        os.chdir(DENO_SCRIPT_PATH)
        deno_process = deno_vendor()
        if deno_process.poll() is None:
            self.stdout.write(f"Starting {deno_vendor}\npid={deno_process.pid}")



import base64
import hashlib

from django.conf import settings

from ..conf.settings import DENO_SERVER, DENO_URL, DENO_TIMEOUT

from .json import JsonApi


class DenoApi(JsonApi):

    server = DENO_SERVER
    url = DENO_URL
    timeout = DENO_TIMEOUT
    use_site_id = True

    def build_post_request_json(self):
        self.request_json = super().build_post_request_json()
        if self.use_site_id:
            site_hash = hashlib.sha256(f"{settings.BASE_DIR}{settings.SESSION_COOKIE_NAME}".encode('utf-8')).digest()
            self.request_json['site_id'] = base64.b64encode(site_hash).decode('utf-8')
        return self.request_json



# https://regexr.com
class MatchGroup(str):
    pass


# regex can be without capturing group.
def finditer_with_separators(regex, s):
    matches = []
    prev_end = 0
    for match in regex.finditer(s):
        match_start = match.start()
        if (prev_end != 0 or match_start > 0) and match_start != prev_end:
            matches.append(s[prev_end:match.start()])
        matches.append(MatchGroup(match.group()))
        prev_end = match.end()
    if prev_end < len(s):
        matches.append(s[prev_end:])
    return matches


# regex should have capturing group.
def split_with_separators(regex, s):
    matches = list(filter(None, regex.split(s)))
    return matches



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Setup ~/.pypirc at https://packaging.python.org/guides/migrating-to-pypi-org/
# python setup.py sdist
# pip3 install twine
# twine upload -r test dist/django-deno-0.1.0.tar.gz

import os
import sys

import django_deno

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

version = django_deno.__version__

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    sys.exit()

if sys.argv[-1] == 'tag':
    print("Tagging the version on github:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

lines = []
with open('README.rst', 'r') as readme_file:
    for line in readme_file:
        # Do not include github relative links which are not parsed by pypi.
        if '.. github relative links' in line:
            break
        else:
            lines.append(line)

readme = ''.join(lines)

# http://stackoverflow.com/questions/14399534/how-can-i-reference-requirements-txt-for-the-install-requires-kwarg-in-setuptool
with open('requirements.txt', 'r') as f:
    install_reqs = [
        s for s in [
            line.split('#', 1)[0].strip(' \t\n') for line in f
        ] if s != ''
    ]

setup(
    name='django-deno',
    version=version,
    description="""Deno front-end integration for Django""",
    long_description=readme,
    long_description_content_type='text/x-rst',
    author='Dmitriy Sintsov',
    author_email='questpc256@gmail.com',
    url='https://github.com/Dmitri-Sintsov/django-deno',
    packages=[
        'django_deno',
    ],
    include_package_data=True,
    install_requires=install_reqs,
    license="LGPL-3.0",
    zip_safe=False,
    keywords='django deno rollup runserver collectstatic'.split(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    setup_requires=['wheel'],
)



import json
import requests
import socket
import time

from jsonschema import validate
from urllib.parse import urlparse
from requests.exceptions import ConnectionError
from urllib3.exceptions import HTTPError


class JsonApi:

    server = None
    url = None
    timeout = None
    request_get_schema = None
    request_post_schema = None
    response_get_schema = None
    response_post_schema = None
    location = '/'
    extra_get_kwargs = {}
    extra_post_kwargs = {}

    scheme_ports = {
        'http': 80,
        'https': 443,
    }

    def __init__(self, **kwargs):
        if self.server is None:
            parts = urlparse(self.url)
            self.server = {
                'scheme': parts.scheme,
                'hostname': parts.hostname,
                'port': self.scheme_ports[parts.scheme] if parts.port is None else parts.port
            }
        elif self.url is None:
            self.url = f'{self.server["scheme"]}://{self.server["hostname"]}:{self.server["port"]}'

    def parse_get_response(self, response):
        response_data = response.json()
        if self.response_get_schema is not None:
            validate(instance=response_data, schema=self.response_get_schema)
        return response_data

    def parse_post_response(self, response):
        response_data = response.json()
        if self.response_post_schema is not None:
            validate(instance=response_data, schema=self.response_post_schema)
        return response_data

    def build_request_json(self):
        return self.request_json

    def build_get_request_json(self):
        return self.build_request_json()

    def build_post_request_json(self):
        return self.build_request_json()

    def build_request_kwargs(self, context):
        request_kwargs = {
            'url': f'{self.url}{self.location}',
        }
        request_json = context['build_request_json']()
        if request_json is not None:
            request_kwargs['json'] = request_json
        request_kwargs.update(context['extra_kwargs'])
        return request_kwargs

    def build_get_kwargs(self):
        return self.build_request_kwargs({
            'build_request_json': self.build_get_request_json,
            'extra_kwargs': self.extra_get_kwargs,
        })

    def build_post_kwargs(self):
        return self.build_request_kwargs({
            'build_request_json': self.build_post_request_json,
            'extra_kwargs': self.extra_post_kwargs,
        })

    def parse_not_responding_error(self, ex):
        # none indicates server is not responding (or not running)
        return None

    def parse_http_error(self, ex):
        return ex

    def set_timeout(self, timeout):
        self.timeout = timeout
        return self

    def get(self, request_json=None):
        context = {
            'requests_method': requests.get,
            'validate_request': self.validate_get_request,
            'build_kwargs': self.build_get_kwargs,
            'parse_response': self.parse_get_response,
        }
        return self.method(context, request_json)

    def post(self, request_json=None):
        context = {
            'requests_method': requests.post,
            'validate_request': self.validate_post_request,
            'build_kwargs': self.build_post_kwargs,
            'parse_response': self.parse_post_response,
        }
        return self.method(context, request_json)

    def validate_get_request(self):
        if self.request_get_schema is not None:
            validate(instance=self.request_json, schema=self.request_get_schema)

    def validate_post_request(self):
        if self.request_post_schema is not None:
            validate(instance=self.request_json, schema=self.request_post_schema)

    def method(self, context, request_json):
        self.request_json = request_json
        context['validate_request']()
        try:
            self.wait_socket(self.timeout)
            request_kwargs = context['build_kwargs']()
            response = context['requests_method'](**request_kwargs)
            try:
                return context['parse_response'](response)
            except json.decoder.JSONDecodeError as ex:
                return HTTPError(f'JSONDecodeError {str(ex)} received text: "{response.text}"')
            except Exception as ex:
                return ex
        except (ConnectionError, TimeoutError) as ex:
            return self.parse_not_responding_error(ex)
        except HTTPError as ex:
            return self.parse_http_error(ex)

    # https://gist.github.com/butla/2d9a4c0f35ea47b7452156c96a4e7b12
    def wait_socket(self, timeout=None):
        """Wait until a port starts accepting TCP connections.
        Args:
            timeout (float): In seconds. How long to wait before raising errors.
        Raises:
            TimeoutError: The port isn't accepting connection after time specified in `timeout`.
        """
        if timeout is not None:
            start_time = time.perf_counter()
            while True:
                try:
                    with socket.create_connection((self.server['hostname'], self.server['port']), timeout=timeout):
                        break
                except OSError as ex:
                    time.sleep(0.01)
                    if time.perf_counter() - start_time >= timeout:
                        raise TimeoutError(
                            f"Waited too long for the port {self.server['port']} on host {self.server['hostname']} "
                            "to start accepting connections."
                        ) from ex



import os
import ijson
import json
import signal


from django.core.cache import cache
from django.contrib.staticfiles.management.commands import collectstatic

from ...itero import IterO
from ...commands import DenoProcess
from ...importmap import ImportMapGenerator
from ...sourcefile import SourceFile
from ...utils import unfold

from ...conf import settings as deno_settings
from ...api.rollup import DenoRollup


class Command(collectstatic.Command, DenoProcess):

    help = f"{collectstatic.Command.help} Uses deno rollup to collect bundles from static files in a single location."

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        import_map_generator = ImportMapGenerator(logger=self.stderr)
        self.serialized_map_generator = json.dumps(import_map_generator.serialize(), separators=(',', ':')).encode('utf-8')
        self.deno_process = self.run_deno_process()
        self.rollup_files = []

    def write_rollup_response(self, objects, prefixed_path):
        for chunks in objects:
            for obj in chunks:
                is_main_chunk = prefixed_path.endswith(obj['filename'])
                dest_js_filename = self.storage.path(prefixed_path)
                dest_dir = os.path.dirname(dest_js_filename)
                if not is_main_chunk:
                    dest_js_filename = dest_dir + os.sep + obj['filename']
                os.makedirs(dest_dir, exist_ok=True)
                dest_map_filename = f"{dest_js_filename}.map"
                with open(dest_js_filename, "w", encoding='utf-8') as f:
                    f.write(obj['code'])
                    f.write(f"//# sourceMappingURL={obj['filename']}.map")
                self.stdout.write(f"Writing '{dest_js_filename}'")
                with open(dest_map_filename, "w", encoding='utf-8') as f:
                    f.write(obj['map'])
                self.stdout.write(f"Writing '{dest_map_filename}'")

    def cache_set(self, source_file, data):
        cache.set(f'deno_rollup_response_{source_file.source_path}', data, timeout=None)

    def cache_get(self, source_file):
        return cache.get(f'deno_rollup_response_{source_file.source_path}')

    def cache_delete(self, source_file):
        cache.delete(f'deno_rollup_response_{source_file.source_path}')

    def cache_rollup_file(self, source_file):
        if not self.is_local_storage():
            self.terminate("Only local storage is supported for rollup files")
        response = DenoRollup(content_type=source_file.content_type).post({
            'filename': str(source_file.source_path.name),
            'basedir': str(source_file.source_path.parent),
            'options': deno_settings.DENO_ROLLUP_COLLECT_OPTIONS,
        })
        if response.status_code == 200 and hasattr(response, 'streaming_content'):
            response_io = IterO(response.streaming_content)
            # s = response_io.read()
            objects = ijson.items(response_io, prefix='rollupFiles')
            cached_objects = unfold(objects, max_level=2)
            self.cache_set(source_file, cached_objects)
        else:
            self.terminate(f"Error: status={response.status_code} content={response.content}")

    def copy_file(self, path, prefixed_path, source_storage):
        source_path = source_storage.path(path)
        source_file = SourceFile(source_path)
        if source_file.should_rollup():
            destination_path = self.storage.path(path)
            self.rollup_files.append([path, prefixed_path, destination_path])
        super().copy_file(path, prefixed_path, source_storage)

    def sigint_handler(self, signum, frame):
        self.stdout.write(f"Terminating deno server pid={self.deno_process.pid}")
        self.deno_process.terminate()
        if callable(self.orig_sigint):
            self.orig_sigint(signum, frame)

    def handle(self, **options):
        self.orig_sigint = signal.getsignal(signal.SIGINT)
        signal.signal(signal.SIGINT, self.sigint_handler)
        super().handle(**options)
        for _path, _prefixed_path, destination_path in self.rollup_files:
            destination_file = SourceFile(destination_path)
            self.cache_rollup_file(destination_file)
        for _path, prefixed_path, destination_path in self.rollup_files:
            destination_file = SourceFile(destination_path)
            cache_entry = self.cache_get(destination_file)
            self.write_rollup_response(cache_entry, prefixed_path)
            self.cache_delete(destination_file)
        if self.is_spawned_deno(deno_process=self.deno_process):
            self.deno_process.terminate()



import collections
import traceback


def ex_to_str(ex):
    return '\n'.join(
        list(traceback.TracebackException.from_exception(ex).format())
    )


def unfold(iterable, max_level=None, curr_level=1):
    ret = []
    for element in iterable:
        if isinstance(element, collections.Iterable) and (max_level is None or curr_level < max_level):
            ret.append(unfold(element, max_level=max_level, curr_level=curr_level + 1))
        else:
            ret.append(element)
    return ret



from .base import ExecDeno


# deno vendor --unstable --force --lock=/home/user/work/djk-sample38/lib/python3.8/site-packages/django_deno/deno/lock.json --import-map=/home/user/work/djk-sample38/lib/python3.8/site-packages/django_deno/deno/import_map.json /home/user/work/djk-sample38/lib/python3.8/site-packages/django_deno/deno/server.ts
# deno run -A --unstable --allow-net --no-remote --import-map vendor/import_map.json /home/user/work/djk-sample38/lib/python3.8/site-packages/django_deno/deno/server.ts --host=127.0.0.1 --port=8099

class DenoVendor(ExecDeno):

    deno_command = 'vendor'
    deno_flags = ["--unstable", "--force"]
    script_name = "server.ts"

    def get_deno_flags(self):
        deno_flags = super().get_deno_flags()
        deno_flags.extend([
            f"--lock={self.deno_lock_path}",
            f"--import-map={self.run_importmap_path}",
        ])
        return deno_flags



import os
import psutil
import subprocess
import _thread

from django_deno import __version__

from .utils import ex_to_str
from .api.maps import DenoMaps
from .process.server import DenoServer
from .importmap import ImportMapGenerator


class DenoProcess:

    def terminate(self, error_message):
        self.stderr.write(error_message)
        # raising CommandError will not shut down test server as it's running in separate thread.
        _thread.interrupt_main()
        # Need to use an OS exit because sys.exit doesn't work in a thread
        os._exit(1)

    def is_spawned_deno(self, deno_process):
        return isinstance(deno_process, subprocess.Popen)

    def is_separate_deno(self, deno_process):
        return isinstance(deno_process, psutil.Process)

    def run_deno_process(self):
        import_map_generator = ImportMapGenerator(logger=self.stderr)
        serialized_map_generator = import_map_generator.serialize()
        deno_api_status = DenoMaps().set_timeout(0.1).post(serialized_map_generator)
        if deno_api_status is None:
            deno_server = DenoServer()
            deno_process = deno_server()
            if deno_process.poll() is None:
                self.stdout.write(f"Starting deno server {deno_server}\npid={deno_process.pid}")
            else:
                self.terminate("Error while starting deno server")
        elif isinstance(deno_api_status, Exception):
            self.stderr.write("The service running is not deno server or deno server is not running properly")
            self.terminate(ex_to_str(deno_api_status))
        else:
            deno_process = psutil.Process(deno_api_status['pid'])
            self.stdout.write(
                f"Already running deno server pid={deno_process.pid}, api version={deno_api_status['version']}"
            )
        if not isinstance(deno_api_status, dict):
            self.stdout.write(f"Sending import maps to deno server pid={deno_process.pid}")
            deno_api_status = DenoMaps().post(serialized_map_generator)
        if isinstance(deno_api_status, Exception) or deno_api_status is None:
            self.stderr.write("The service running is not deno server or deno server is not running properly")
            if isinstance(deno_api_status, Exception):
                self.terminate(ex_to_str(deno_api_status))
            else:
                self.terminate('')
        elif deno_api_status['version'] != __version__:
            self.terminate(
                f"Version of deno server api does not match, "
                f"found: {deno_api_status['version']}, required: {__version__}"
            )
        self.stdout.write(f"Sent import maps to deno server pid={deno_process.pid}")
        return deno_process
